#! -*- coding: utf8 -*-

'''
Se mete entre el cliente y el server X. En un futuro permitirá
redireccionar este flujo.

Ejemplo:

  Ejecutar en una terminal: python xred.py 6010
  En otra terminal: DISPLAY=localhost:10 xterm

TODO:
 - Que al ejecutarlo sin parametros informe si hay algún otro xred corriendo,
   y en que puerto está escuchando
 - Que el manejo de cada programa lo haga en un hilo separado, así puede servir a
   muchos en el mismo puerto
 - Que se ponga a escuchar en un puerto alto mensajes de control.
 - Que al ejecutarlo guarde en algo como ~/.xred a quien está sirviendo, y en que
   puerto escucha mensajes de control.
 - Que si recibe un mensaje de control determinado, switchee a trabajar con otro
   servidor X
 - Que soporte ser llamado "xred --tomar=id", donde id es el mismo que muestra si
   lo llamamos sin parametros. Al ser llamado con "tomar", le manda un mensaje de
   control a ese que ya está corriendo para que switchee el server.
'''





USAGE = """Usar:
  xmman.py help        Este mensaje
  xmman.py start       Arranca el demonio
  xmman.py serve 1234  Pone un XMiddleMan a servir en el puerto 1234
  xmman.py list        Lista las conexiones actuales
  xmman.py status      Muestra el estado del demonio XMMan
"""

import socket, select, sys, threading
import SimpleXMLRPCServer
import xmlrpclib, user, os

class XMLRPCServerInterrumpible(SimpleXMLRPCServer.SimpleXMLRPCServer):
    def __init__(self, hostport):
        self.servir = True
        SimpleXMLRPCServer.SimpleXMLRPCServer.__init__(self, hostport)

    def serve_forever(self):
        """Redefinido para cortar en función de self.servir."""
        while self.servir:
            self.handle_request()

    def server_close(self):
        """Limpia y para el server."""
        self.socket.close()
        self.servir = False

class XHandler(threading.Thread):
    '''Implementamos la conexión en si.'''
    def __init__(self, puerto, nosfuimos):
        # nos conectamos al servidor X local
        print "XHandler: conectando al server X"
        self.xserver =  socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.xserver.connect("/tmp/.X11-unix/X0")
        self.puerto = puerto

        # nos ponemos a servir
        print "Main: sirviendo en el puerto", puerto
        self.nos = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.nos.bind(("localhost", puerto))
        self.nos.listen(1)

        self.accepted = False
        self.nosfuimos = nosfuimos
        self.servir = True
        threading.Thread.__init__(self)

    def stop(self):
        self.servir = False

        # si no tuvimos conexión,, nos conectamos a nosotros mismos para destrabarnos
        if not self.accepted:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect(("localhost", self.puerto))

    def run(self):
        xserver = self.xserver
        # esperamos una conexión
        xclient, addr = self.nos.accept()
        self.accepted = True
        print "Main: conectado desde", addr

        # cruzamos la info para un lado y para el otro
        while self.servir:
            (leer, escr, errr) = select.select([xclient, xserver], [], [], 1)
            if xclient in leer:
                data = xclient.recv(4)
                if not data:
                    break
                print "client -> server", repr(data)
                xserver.send(data)
            if xserver in leer:
                data = xserver.recv(4)
                print "server -> client", repr(data)
                xclient.send(data)
        self.nosfuimos()


class Director(threading.Thread):
    '''Supervisora del funcionamiento en general.

    Esta clase provee las distintas funciones para manejar las distintas
    conexiones levantadas.
    '''
    def __init__(self):
        self.activeHandlers = set()
        threading.Thread.__init__(self)
        self.xhandlers = []

    def newXHandler(self, puerto):
        '''Levanta un nuevo handler en el puerto indicado.'''
        if puerto in self.activeHandlers:
            raise ValueError("Puerto %r ya tomado por otro handler!" % puerto)

        self.activeHandlers.add(puerto)
        def f():
            self.activeHandlers.remove(puerto)
        xh = XHandler(puerto, f)
        xh.start()
        self.xhandlers.append(xh)
        return ""

    def run(self):
        '''Nos ponemos a servir XMLRPC.'''
        self.server = XMLRPCServerInterrumpible(("localhost", 5999))
        self.server.register_function(self.list)
        self.server.register_function(self.stop)
        self.server.register_function(self.newXHandler)
        self.server.serve_forever()
        print "FNISSSSSSSSSSS"

    def list(self):
        '''Devuelve las conexiones activas.'''
        return list(self.activeHandlers)

    def stop(self):
        '''Cierra el server.'''
        self.server.server_close()
        for xh in self.xhandlers:
            xh.stop()
        return ""

class Cliente(object):
    '''Implementa todas las preguntas que se le pueden hacer al Director.'''
    def __init__(self):
        # revisamos si hay alguna instancia trabajando
        self.ourdir = os.path.join(user.home, ".xmman")
        self.yaExiste = self._controlInstancia()
        self.server = xmlrpclib.Server("http://localhost:5999")

    def _controlInstancia(self):
        '''Devuelve si ya hay una instancia del server.'''
        lockfile = os.path.join(self.ourdir, "pid.lock")
        return os.access(lockfile, os.F_OK)

    def help(self, *args):
        print USAGE

    def start(self):
        if self.yaExiste:
            print "Ya hay un demonio XMMan activo!"
            return

        # creamos el dir si lo necesitamos
        if not os.access(self.ourdir, os.F_OK):
            os.mkdir(self.ourdir)

        # marcamos el lock
        lockfile = os.path.join(self.ourdir, "pid.lock")
        a = open(lockfile, "w")
        a.write("%s\n" % os.getpid())
        a.close()

        # creamos el servicio
        self.yaExiste = True
        dir = Director()
        dir.start()


    def serve(self, port):
        if not port.isdigit():
            raise TypeError("El puerto tiene que ser numérico!")
        port = int(port)

        if not self.yaExiste:
            print "No hay un demonio XMMan activo!"
            return

        self.server.newXHandler(port)

    def stop(self):
        # FIXME agregar
        self.server.stop()
        lockfile = os.path.join(self.ourdir, "pid.lock")
        os.remove(lockfile)

    def status(self):
        '''Muestra el estado del server.'''
        lockfile = os.path.join(self.ourdir, "pid.lock")
        if os.access(lockfile, os.F_OK):
            pid = open(lockfile).read().strip()
            print "Hay un demonio activo, process id", pid
        else:
            print "No hay un demonio XMMan activo."

    def list(self):
        '''Muestra las conexiones activas.'''
        print self.server.list()

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print USAGE
        sys.exit(-1)

    cliente = Cliente()

    try:
        func = getattr(cliente, sys.argv[1])
    except AttributeError:
        print "Error! Orden no encontrada"
        print USAGE
        sys.exit(-1)

    try:
        func(*sys.argv[2:])
    except TypeError, e:
        print "Error!", str(e)
        print USAGE
        sys.exit(-1)
    except xmlrpclib.ProtocolError, e:
        print "Error!", str(e)
        print 'Pruebe con "xmman.py status" para saber el estado del demonio XMMan'
        sys.exit(-2)
